@@ -1,3 +1,13 @@
+# vim: set fileencoding=utf-8:
+#
+# GPIO Zero: a library for controlling the Raspberry Pi's GPIO pins
+#
+# Copyright (c) 2016-2021 Dave Jones <dave@waveform.org.uk>
+# Copyright (c) 2019 Ben Nuttall <ben@bennuttall.com>
+# Copyright (c) 2018 Martchus <martchus@gmx.net>
+#
+# SPDX-License-Identifier: BSD-3-Clause
+
 from __future__ import (
     unicode_literals,
     absolute_import,
@@ -7,7 +17,9 @@ from __future__ import (
 str = type('')
 
 
+import io
 import re
+import errno
 import pytest
 from mock import patch, MagicMock
 
@@ -21,29 +33,66 @@ from gpiozero import *
 def test_pi_revision():
     with patch('gpiozero.devices.Device.pin_factory', LocalPiFactory()):
         # Can't use MockPin for this as we want something that'll actually try
-        # and read /proc/cpuinfo (MockPin simply parrots the 2B's data);
-        # LocalPiFactory is used as we can definitely instantiate it (strictly
-        # speaking it's abstract but we're only interested in the pi_info
-        # stuff)
+        # and read /proc/device-tree/system/linux,revision and /proc/cpuinfo
+        # (MockPin simply parrots the 3B's data); LocalPiFactory is used as we
+        # can definitely instantiate it (strictly speaking it's abstract but
+        # we're only interested in the pi_info stuff)
         with patch('io.open') as m:
-            m.return_value.__enter__.return_value = ['lots of irrelevant', 'lines', 'followed by', 'Revision: 0002', 'Serial:  xxxxxxxxxxx']
+            m.return_value.__enter__.side_effect = [
+                # Pretend /proc/device-tree/system/linux,revision doesn't
+                # exist, and that /proc/cpuinfo contains the Revision: 0002
+                # after some filler
+                IOError(errno.ENOENT, 'File not found'),
+                ['lots of irrelevant', 'lines', 'followed by', 'Revision: 0002', 'Serial:  xxxxxxxxxxx']
+            ]
             assert pi_info().revision == '0002'
             # LocalPiFactory caches the revision (because realistically it
             # isn't going to change at runtime); we need to wipe it here though
             Device.pin_factory._info = None
-            m.return_value.__enter__.return_value = ['Revision: a21042']
+            m.return_value.__enter__.side_effect = [
+                IOError(errno.ENOENT, 'File not found'),
+                ['Revision: a21042']
+            ]
             assert pi_info().revision == 'a21042'
-            # Check over-volting result (some argument over whether this is 7 or
-            # 8 character result; make sure both work)
+            # Check over-volting result (some argument over whether this is 7
+            # or 8 character result; make sure both work)
             Device.pin_factory._info = None
-            m.return_value.__enter__.return_value = ['Revision: 1000003']
+            m.return_value.__enter__.side_effect = [
+                IOError(errno.ENOENT, 'File not found'),
+                ['Revision: 1000003']
+            ]
             assert pi_info().revision == '0003'
             Device.pin_factory._info = None
-            m.return_value.__enter__.return_value = ['Revision: 100003']
+            m.return_value.__enter__.side_effect = [
+                IOError(errno.ENOENT, 'File not found'),
+                ['Revision: 100003']
+            ]
             assert pi_info().revision == '0003'
+            # Check we complain loudly if we can't access linux,revision
+            Device.pin_factory._info = None
+            m.return_value.__enter__.side_effect = [
+                # Pretend /proc/device-tree/system/linux,revision doesn't
+                # exist, and that /proc/cpuinfo contains the Revision: 0002
+                # after some filler
+                IOError(errno.EACCES, 'Permission denied'),
+                ['Revision: 100003']
+            ]
+            with pytest.raises(IOError):
+                pi_info()
+            # Check that parsing /proc/device-tree/system/linux,revision also
+            # works properly
+            Device.pin_factory._info = None
+            m.return_value.__enter__.side_effect = None
+            m.return_value.__enter__.return_value = io.BytesIO(b'\x00\xa2\x20\xd3')
+            assert pi_info().revision == 'a220d3'
+            # Check that if everything's a bust we raise PinUnknownPi
             with pytest.raises(PinUnknownPi):
-                m.return_value.__enter__.return_value = ['nothing', 'relevant', 'at all']
                 Device.pin_factory._info = None
+                m.return_value.__enter__.return_value = None
+                m.return_value.__enter__.side_effect = [
+                    IOError(errno.ENOENT, 'File not found'),
+                    ['nothing', 'relevant']
+                ]
                 pi_info()
             with pytest.raises(PinUnknownPi):
                 pi_info('0fff')
@@ -73,6 +122,16 @@ def test_pi_info():
     assert not r.bluetooth
     assert r.csi == 1
     assert r.dsi == 1
+    assert repr(r).startswith('PiBoardInfo(revision=')
+    assert 'headers=...' in repr(r)
+
+def test_pi_info_not_a_pi():
+    class NotAPiFactory(LocalPiFactory):
+        def _get_pi_info(self):
+            return None
+    with patch('gpiozero.devices.Device.pin_factory', NotAPiFactory()):
+        with pytest.raises(PinUnknownPi):
+            pi_info()
 
 def test_pi_info_other_types():
     assert pi_info(b'9000f1') == pi_info(0x9000f1)
@@ -131,6 +190,8 @@ def test_format_content():
         pi_info('900092').pprint(color=True)
         s = ''.join(stdout.output)
         assert '{0:color full}\n'.format(pi_info('900092')) == s
+        with pytest.raises(ValueError):
+            '{0:color foo}'.format(pi_info('900092'))
 
 def test_pprint_headers():
     assert len(pi_info('0002').headers) == 1
@@ -155,6 +216,21 @@ def test_pprint_headers():
         assert 'P1:\n' not in s
         assert 'P5:\n' not in s
 
+def test_format_headers():
+    with patch('sys.stdout') as stdout:
+        stdout.output = []
+        stdout.write = lambda buf: stdout.output.append(buf)
+        info = pi_info('c03131')
+        info.headers['J8'].pprint(color=False)
+        s = ''.join(stdout.output)
+        assert '{0.headers[J8]:mono}\n'.format(info) == s
+        stdout.output = []
+        info.headers['J8'].pprint(color=True)
+        s = ''.join(stdout.output)
+        assert '{0.headers[J8]:color}\n'.format(info) == s
+        with pytest.raises(ValueError):
+            '{0.headers[J8]:mono foo}'.format(info)
+
 def test_pprint_color():
     with patch('sys.stdout') as stdout:
         stdout.output = []
@@ -221,4 +297,3 @@ def test_pprint_rows_cols():
         '{0:row16}'.format(pi_info('0002').headers['P1'])
     with pytest.raises(ValueError):
         '{0:col3}'.format(pi_info('0002').headers['P1'])
-
