@@ -1,3 +1,13 @@
+# vim: set fileencoding=utf-8:
+#
+# GPIO Zero: a library for controlling the Raspberry Pi's GPIO pins
+#
+# Copyright (c) 2016-2021 Dave Jones <dave@waveform.org.uk>
+# Copyright (c) 2018-2019 Ben Nuttall <ben@bennuttall.com>
+# Copyright (c) 2016-2019 Andrew Scheller <github@loowis.durge.org>
+#
+# SPDX-License-Identifier: BSD-3-Clause
+
 from __future__ import (
     unicode_literals,
     absolute_import,
@@ -9,9 +19,10 @@ str = type('')
 
 import pytest
 from math import sin, cos, radians
-from time import time
+from time import time, sleep
 from itertools import islice
 
+from gpiozero import Device, LED, Button, Robot
 from gpiozero.tools import *
 try:
     from math import isclose
@@ -27,10 +38,68 @@ except ImportError:
     from gpiozero.compat import median
 
 
-def test_negated():
+epsilon = 0.01  # time to sleep after setting source before checking value
+
+def test_set_source_by_value(mock_factory):
+    with LED(2) as led, Button(3) as btn:
+        led.source_delay = 0
+        assert not led.value
+        assert not btn.value
+        led.source = btn.values
+        sleep(epsilon)
+        assert not led.value
+        assert not btn.value
+        btn.pin.drive_low()
+        sleep(epsilon)
+        assert led.value
+        assert btn.value
+
+def test_set_source_by_device(mock_factory):
+    with LED(2) as led, Button(3) as btn:
+        led.source_delay = 0
+        assert not led.value
+        assert not btn.value
+        led.source = btn
+        sleep(epsilon)
+        assert not led.value
+        assert not btn.value
+        btn.pin.drive_low()
+        sleep(epsilon)
+        assert led.value
+        assert btn.value
+
+def test_negated(mock_factory):
     assert list(negated(())) == []
     assert list(negated((True, True, False, False))) == [False, False, True, True]
 
+def test_negated_source_by_value(mock_factory):
+    with LED(2) as led, Button(3) as btn:
+        led.source_delay = 0
+        assert not led.value
+        assert not btn.value
+        led.source = negated(btn.values)
+        sleep(epsilon)
+        assert led.value
+        assert not btn.value
+        btn.pin.drive_low()
+        sleep(epsilon)
+        assert not led.value
+        assert btn.value
+
+def test_negated_source_by_device(mock_factory):
+    with LED(2) as led, Button(3) as btn:
+        led.source_delay = 0
+        assert not led.value
+        assert not btn.value
+        led.source = negated(btn)
+        sleep(epsilon)
+        assert led.value
+        assert not btn.value
+        btn.pin.drive_low()
+        sleep(epsilon)
+        assert not led.value
+        assert btn.value
+
 def test_inverted():
     with pytest.raises(ValueError):
         list(inverted((), 0, 0))
@@ -74,6 +143,12 @@ def test_scaled():
     # scale -1->+1 to 0->1
     assert list(scaled((-1, 1, 0.0, -0.5), 0, 1, -1, 1)) == [0, 1, 0.5, 0.25]
 
+def test_scaled_full():
+    assert list(scaled_full((0, 0.5, 1, 0.5))) == [-1, 0, 1, 0]
+
+def test_scaled_half():
+    assert list(scaled_half((-1, 0, 1, 0))) == [0, 0.5, 1, 0.5]
+
 def test_clamped():
     with pytest.raises(ValueError):
         list(clamped((), 0, 0))
@@ -87,7 +162,7 @@ def test_clamped():
     assert list(clamped((-2, -1, -0.5, 0, 0.5, 1, 2), -1, 1)) == [-1, -1, -0.5, 0, 0.5, 1, 1]
     assert list(clamped((-2, -1, -0.5, 0, 0.5, 1, 2), -2, 2)) == [-2, -1, -0.5, 0, 0.5, 1, 2]
 
-def test_absoluted():
+def test_absoluted(mock_factory):
     assert list(absoluted(())) == []
     assert list(absoluted((-2, -1, 0, 1, 2))) == [2, 1, 0, 1, 2]
 
@@ -296,6 +371,7 @@ def test_cos_values():
 
 def test_ramping_values():
     assert list(islice(ramping_values(2), 2)) == [0, 1]
+    assert list(islice(ramping_values(3), 5)) == [0, 2/3, 2/3, 0, 2/3]
     assert list(islice(ramping_values(4), 4)) == [0, 0.5, 1, 0.5]
     assert list(islice(ramping_values(8), 8)) == [0, 0.25, 0.5, 0.75, 1, 0.75, 0.5, 0.25]
     firstval = None
@@ -315,3 +391,31 @@ def test_ramping_values():
             else:
                 if i % period == 0:
                     assert v == firstval
+
+def test_zip_values(mock_factory):
+    with Button(2) as btn1, Button(3) as btn2:
+        zv = zip_values(btn1, btn2)
+        assert next(zv) == (False, False)
+        btn1.pin.drive_low()
+        assert next(zv) == (True, False)
+        btn2.pin.drive_low()
+        assert next(zv) == (True, True)
+        btn1.pin.drive_high()
+        assert next(zv) == (False, True)
+        btn2.pin.drive_high()
+        assert next(zv) == (False, False)
+    with Button(2) as btn1, Button(3) as btn2, Button(4) as btn3, Button(5) as btn4:
+        zv = zip_values(btn1, btn2, btn3, btn4)
+        assert next(zv) == (False, False, False, False)
+        btn1.pin.drive_low()
+        btn3.pin.drive_low()
+        assert next(zv) == (True, False, True, False)
+        btn2.pin.drive_low()
+        btn4.pin.drive_low()
+        assert next(zv) == (True, True, True, True)
+        btn1.pin.drive_high()
+        btn2.pin.drive_high()
+        btn3.pin.drive_high()
+        btn4.pin.drive_high()
+        sleep(epsilon)
+        assert next(zv) == (False, False, False, False)
